/*
 * Hash Table API Test.
 *
 * Compile with:
 * 	gcc -I../lib -I../include -L../lib -O0 -Wall -o hash_test hash_test.c -llib
 */
#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>

#include "macros.h"
#include "hash.h"

#define ERROR(fmt, arg...) \
do { \
	fprintf(stderr, "hash_table: " fmt "\n", ##arg); \
} while (0)

void
dump_hash(struct hash_table *table)
{
	struct hash_table_iter *iter;
	void *key, *data;
	
	if (table == NULL)
		return;

	iter = hash_table_iterate_init(table);

	printf("=== Hash Table Dump ===\n");

	while (hash_table_iterate(iter, &key, &data)) {
		printf("key %-8lu data %-8lu\n",
		       (unsigned long) key,
		       (unsigned long) data);
	}

	hash_table_iterate_deinit(&iter);
}

int
hash_test()
{
	int res, i;
	struct hash_table *table;
	struct hash_table_iter *iter;
	void *kp, *dp;
	unsigned long keys[] = {
		12, 43, 98,   123, 1923, 7732, 553,  23992,
		94, 78, 8472, 712, 6453, 9818, 1097, 89
	};
	unsigned long data[] = {
		0, 1,  2,  3,  4,  5,  6, 7,
		8, 9, 10, 11, 12, 13, 14, 15,
	};

#define NKEYS	(sizeof(keys)/sizeof(unsigned long))

	table = hash_table_new(3, NULL, NULL);

	if (table == NULL) {
		fprintf(stderr, "hash_table: can't create hash table\n");
		return 1;
	}

	/*
	 * Insert and remove test.
	 */

	for (i = 0; i < NKEYS; i++) {
		res = hash_table_insert(table, (void *)keys[i], (void *)data[i]);
		if (res != ret_ok) {
			ERROR("failed to add %d-th key to hash table", i);
			goto error;
		}
	}

	for (i = 0; i < NKEYS; i++) {
		res = hash_table_remove(table, (void *)keys[i]);
		if (res != TRUE) {
			ERROR("failed to remove %d-th key from hash table", i);
			goto error;
		}
	}

	hash_table_clean(table);

	/*
	 * Insert unique test.
	 */
	
	for (i = 0; i < NKEYS; i++) {
		res = hash_table_insert_unique(table, (void *)keys[i], (void *)data[i]);
		if (res != ret_ok) {
			ERROR("failed to add %d-th key to hash table at second turn", i);
			goto error;
		}
	}

	/* Check that all keys are present in the table. */
	iter = hash_table_iterate_init(table);

	while (hash_table_iterate(iter, &kp, &dp) != 0) {
		int idx = (unsigned long) dp;
		if (idx >= NKEYS) {
			ERROR("data pointer overwritten in hash table");
			goto error;
		}
		if ((void *)keys[idx] != kp) {
			ERROR("key pointer overwritten in hash table");
			goto error;
		}
	}

	hash_table_iterate_deinit(&iter);

	for (i = 0; i < NKEYS; i++) {
		res = hash_table_insert_unique(table, (void *)keys[i], (void *)data[i]);
		if (res != ret_entry_exists) {
			ERROR("no %d-th unique key present in table after insert", i);
			goto error;
		}
	}

	/*
	 * Replace test.
	 */
	
	for (i = 0; i < NKEYS; i++) {
		res = hash_table_replace(table, (void*)keys[i], (void*)data[NKEYS-i-1]);
		if (res != TRUE) {
			ERROR("no %d-th key present in table during replace", i);
			goto error;
		}
	}

	iter = hash_table_iterate_init(table);

	while (hash_table_iterate(iter, &kp, &dp) != 0) {
		int idx = NKEYS - (unsigned long)dp - 1;
		if (idx >= NKEYS) {
			ERROR("data pointer overwritten in hash table after replace");
			goto error;
		}
		if ((void *)keys[idx] != kp) {
			ERROR("key pointer overwritten in hash table after replace");
			goto error;
		}
	}

	hash_table_iterate_deinit(&iter);

	for (i = 0; i < NKEYS; i++) {
		res = hash_table_lookup(table, (void*)keys[i], NULL);
		if (res == ret_not_found) {
			ERROR("no %d-th key present in table after replace", i);
			goto error;
		}
	}

	hash_table_destroy(&table);
	return 0;

error:
	dump_hash(table);
	return 1;
}

int
main(int argc, char *argv[])
{
	int res;
	char *msg[] = { "ok", "failed" };

	res = hash_test();

	fprintf(stderr, "hash_table: %s\n", msg[res]);

	return res;
}
